"""Module that contains the classes for actions and workflows."""

import traceback
from abc import abstractmethod
from asyncio import Queue, create_task
from typing import Any, Dict, Self, Tuple, Type, Union, Unpack, final

from fabricatio.capabilities.rating import GiveRating
from fabricatio.capabilities.task import HandleTask, ProposeTask
from fabricatio.journal import logger
from fabricatio.models.generic import WithBriefing
from fabricatio.models.task import Task
from fabricatio.models.usages import ToolBoxUsage
from pydantic import Field, PrivateAttr


class Action(HandleTask, ProposeTask, GiveRating):
    """Class that represents an action to be executed in a workflow."""

    name: str = Field(default="")
    """The name of the action."""
    description: str = Field(default="")
    """The description of the action."""
    personality: str = Field(default="")
    """The personality of whom the action belongs to."""
    output_key: str = Field(default="")
    """The key of the output data."""

    @final
    def model_post_init(self, __context: Any) -> None:
        """Initialize the action by setting the name if not provided.

        Args:
            __context: The context to be used for initialization.
        """
        self.name = self.name or self.__class__.__name__
        self.description = self.description or self.__class__.__doc__ or ""

    @abstractmethod
    async def _execute(self, **cxt: Unpack) -> Any:
        """Execute the action with the provided arguments.

        Args:
            **cxt: The context dictionary containing input and output data.

        Returns:
            The result of the action execution.
        """
        pass

    @final
    async def act(self, cxt: Dict[str, Any]) -> Dict[str, Any]:
        """Perform the action by executing it and setting the output data.

        Args:
            cxt: The context dictionary containing input and output data.
        """
        ret = await self._execute(**cxt)
        if self.output_key:
            logger.debug(f"Setting output: {self.output_key}")
            cxt[self.output_key] = ret
        return cxt

    @property
    def briefing(self) -> str:
        """Return a brief description of the action."""
        if self.personality:
            return f"## Your personality: \n{self.personality}\n# The action you are going to perform: \n{super().briefing}"
        return f"# The action you are going to perform: \n{super().briefing}"


class WorkFlow(WithBriefing, ToolBoxUsage):
    """Class that represents a workflow to be executed in a task."""

    _context: Queue[Dict[str, Any]] = PrivateAttr(default_factory=lambda: Queue(maxsize=1))
    """ The context dictionary to be used for workflow execution."""

    _instances: Tuple[Action, ...] = PrivateAttr(...)
    """ The instances of the workflow steps."""

    steps: Tuple[Union[Type[Action], Action], ...] = Field(...)
    """ The steps to be executed in the workflow, actions or action classes."""
    task_input_key: str = Field(default="task_input")
    """ The key of the task input data."""
    task_output_key: str = Field(default="task_output")
    """ The key of the task output data."""
    extra_init_context: Dict[str, Any] = Field(default_factory=dict, frozen=True)
    """ The extra context dictionary to be used for workflow initialization."""

    def model_post_init(self, __context: Any) -> None:
        """Initialize the workflow by setting fallbacks for each step.

        Args:
            __context: The context to be used for initialization.
        """
        temp = []
        for step in self.steps:
            temp.append(step if isinstance(step, Action) else step())
        self._instances = tuple(temp)

    def inject_personality(self, personality: str) -> Self:
        """Inject the personality of the workflow.

        Args:
            personality: The personality to be injected.

        Returns:
            Self: The instance of the workflow with the injected personality.
        """
        for a in filter(lambda action: not action.personality, self._instances):
            a.personality = personality
        return self

    async def serve(self, task: Task) -> None:
        """Serve the task by executing the workflow steps.

        Args:
            task: The task to be served.
        """
        await task.start()
        await self._init_context(task)
        current_action = None
        try:
            for step in self._instances:
                logger.debug(f"Executing step: {step.name}")
                act_task = create_task(step.act(await self._context.get()))
                if task.is_cancelled():
                    act_task.cancel(f"Cancelled by task: {task.name}")
                    break
                modified_ctx = await act_task
                await self._context.put(modified_ctx)
                current_action = step.name
            logger.info(f"Finished executing workflow: {self.name}")
            final_ctx = await self._context.get()
            if self.task_output_key not in final_ctx:
                logger.warning(
                    f"Task output key: {self.task_output_key} not found in the context, None will be returned. You can check if `Action.output_key` is set the same as `WorkFlow.task_output_key`."
                )

            await task.finish(final_ctx.get(self.task_output_key, None))
        except RuntimeError as e:
            logger.error(f"Error during task: {current_action} execution: {e}")  # Log the exception
            logger.error(traceback.format_exc())  # Add this line to log the traceback
            await task.fail()  # Mark the task as failed

    async def _init_context[T](self, task: Task[T]) -> None:
        """Initialize the context dictionary for workflow execution."""
        logger.debug(f"Initializing context for workflow: {self.name}")
        await self._context.put({self.task_input_key: task, **dict(self.extra_init_context)})

    def steps_fallback_to_self(self) -> Self:
        """Set the fallback for each step to the workflow itself."""
        self.hold_to(self._instances)
        return self

    def steps_supply_tools_from_self(self) -> Self:
        """Supply the tools from the workflow to each step."""
        self.provide_tools_to(self._instances)
        return self



"""A module for defining tools and toolboxes."""

from importlib.machinery import ModuleSpec
from importlib.util import module_from_spec
from inspect import iscoroutinefunction, signature
from types import CodeType, ModuleType
from typing import Any, Callable, Dict, List, Optional, Self, overload

from fabricatio.config import configs
from fabricatio.decorators import logging_execution_info, use_temp_module
from fabricatio.journal import logger
from fabricatio.models.generic import WithBriefing
from pydantic import BaseModel, ConfigDict, Field


class Tool[**P, R](WithBriefing):
    """A class representing a tool with a callable source function."""

    name: str = Field(default="")
    """The name of the tool."""

    description: str = Field(default="")
    """The description of the tool."""

    source: Callable[P, R]
    """The source function of the tool."""

    def model_post_init(self, __context: Any) -> None:
        """Initialize the tool with a name and a source function."""
        self.name = self.name or self.source.__name__

        if not self.name:
            raise RuntimeError("The tool must have a source function.")

        self.description = self.description or self.source.__doc__ or ""
        self.description = self.description.strip()

    def invoke(self, *args: P.args, **kwargs: P.kwargs) -> R:
        """Invoke the tool's source function with the provided arguments."""
        logger.info(f"Invoking tool: {self.name}")
        return self.source(*args, **kwargs)

    @property
    def briefing(self) -> str:
        """Return a brief description of the tool.

        Returns:
            str: A brief description of the tool.
        """
        # 获取源函数的返回类型

        return f"{'async ' if iscoroutinefunction(self.source) else ''}def {self.name}{signature(self.source)}\n{_desc_wrapper(self.description)}"


def _desc_wrapper(desc: str) -> str:
    lines = desc.split("\n")
    lines_indent = [f"    {line}" for line in ['"""', *lines, '"""']]
    return "\n".join(lines_indent)


class ToolBox(WithBriefing):
    """A class representing a collection of tools."""

    tools: List[Tool] = Field(default_factory=list, frozen=True)
    """A list of tools in the toolbox."""

    def collect_tool[**P, R](self, func: Callable[P, R]) -> Callable[P, R]:
        """Add a callable function to the toolbox as a tool.

        Args:
            func (Callable[P, R]): The function to be added as a tool.

        Returns:
            Callable[P, R]: The added function.
        """
        self.tools.append(Tool(source=func))
        return func

    def add_tool[**P, R](self, func: Callable[P, R]) -> Self:
        """Add a callable function to the toolbox as a tool.

        Args:
            func (Callable): The function to be added as a tool.

        Returns:
            Self: The current instance of the toolbox.
        """
        self.collect_tool(logging_execution_info(func))
        return self

    @property
    def briefing(self) -> str:
        """Return a brief description of the toolbox.

        Returns:
            str: A brief description of the toolbox.
        """
        list_out = "\n\n".join([f"{tool.briefing}" for tool in self.tools])
        toc = f"## {self.name}: {self.description}\n## {len(self.tools)} tools available:"
        return f"{toc}\n\n{list_out}"

    def get[**P, R](self, name: str) -> Tool[P, R]:
        """Invoke a tool by name with the provided arguments.

        Args:
            name (str): The name of the tool to invoke.

        Returns:
            Tool: The tool instance with the specified name.

        Raises:
            ValueError: If no tool with the specified name is found.
        """
        tool = next((tool for tool in self.tools if tool.name == name), None)
        if tool is None:
            err = f"No tool with the name {name} found in the toolbox."
            logger.error(err)
            raise ValueError(err)

        return tool

    def __hash__(self) -> int:
        """Return a hash of the toolbox based on its briefing."""
        return hash(self.briefing)


class ToolExecutor(BaseModel):
    """A class representing a tool executor with a sequence of tools to execute."""

    model_config = ConfigDict(use_attribute_docstrings=True)
    candidates: List[Tool] = Field(default_factory=list, frozen=True)
    """The sequence of tools to execute."""

    data: Dict[str, Any] = Field(default_factory=dict)
    """The data that could be used when invoking the tools."""

    def inject_tools[M: ModuleType](self, module: Optional[M] = None) -> M:
        """Inject the tools into the provided module or default."""
        module = module or module_from_spec(spec=ModuleSpec(name=configs.toolbox.tool_module_name, loader=None))
        for tool in self.candidates:
            logger.debug(f"Injecting tool: {tool.name}")
            setattr(module, tool.name, tool.invoke)
        return module

    def inject_data[M: ModuleType](self, module: Optional[M] = None) -> M:
        """Inject the data into the provided module or default."""
        module = module or module_from_spec(spec=ModuleSpec(name=configs.toolbox.data_module_name, loader=None))
        for key, value in self.data.items():
            logger.debug(f"Injecting data: {key}")
            setattr(module, key, value)
        return module

    def execute[C: Dict[str, Any]](self, source: CodeType, cxt: Optional[C] = None) -> C:
        """Execute the sequence of tools with the provided context."""
        cxt = cxt or {}

        @use_temp_module([self.inject_data(), self.inject_tools()])
        def _exec() -> None:
            exec(source, cxt)  # noqa: S102

        _exec()
        return cxt

    @overload
    def take[C: Dict[str, Any]](self, keys: List[str], source: CodeType, cxt: Optional[C] = None) -> C:
        """Check the output of the tools with the provided context."""
        ...

    @overload
    def take[C: Dict[str, Any]](self, keys: str, source: CodeType, cxt: Optional[C] = None) -> Any:
        """Check the output of the tools with the provided context."""
        ...

    def take[C: Dict[str, Any]](self, keys: List[str] | str, source: CodeType, cxt: Optional[C] = None) -> C | Any:
        """Check the output of the tools with the provided context."""
        cxt = self.execute(source, cxt)
        if isinstance(keys, str):
            return cxt[keys]
        return {key: cxt[key] for key in keys}

    @classmethod
    def from_recipe(cls, recipe: List[str], toolboxes: List[ToolBox]) -> Self:
        """Create a tool executor from a recipe and a list of toolboxes."""
        tools = []
        while tool_name := recipe.pop(0):
            for toolbox in toolboxes:
                tools.append(toolbox[tool_name])

        return cls(candidates=tools)



"""Extra models for built-in actions."""

from typing import List

from fabricatio.models.generic import Base, Display, FinalizedDumpAble, PrepareVectorization, ProposedAble
from pydantic import Field


class Equation(Base):
    """Structured representation of mathematical equations (including their physical or conceptual meanings)."""

    description: str = Field(...)
    """A concise explanation of the equation's meaning, purpose, and relevance in the context of the research."""

    latex_code: str = Field(...)
    """The LaTeX code used to represent the equation in a publication-ready format."""


class Figure(Base):
    """Structured representation of figures (including their academic significance and explanatory captions)."""

    description: str = Field(...)
    """A detailed explanation of the figure's content and its role in conveying key insights."""

    figure_caption: str = Field(...)
    """The caption accompanying the figure, summarizing its main points and academic value."""

    figure_path: str = Field(...)
    """The file path to the figure"""


class Highlightings(Base):
    """Structured representation of highlighted elements in an academic paper (including equations, algorithms, figures, and tables)."""

    # Academic Achievements Showcase
    highlighted_equations: List[Equation] = Field(default_factory=list)
    """Core mathematical equations that represent breakthroughs in the field, accompanied by explanations of their physical or conceptual significance."""

    highlighted_algorithms: List[str] = Field(default_factory=list)
    """Pseudocode for key algorithms, annotated to highlight innovative components."""

    highlighted_figures: List[Figure] = Field(default_factory=list)
    """Critical diagrams or illustrations, each accompanied by a caption explaining their academic importance."""

    highlighted_tables: List[str] = Field(default_factory=list)
    """Important data tables, annotated to indicate statistical significance or other notable findings."""


class ArticleEssence(ProposedAble, Display, PrepareVectorization):
    """Structured representation of the core elements of an academic paper(providing a comprehensive digital profile of the paper's essential information)."""

    # Basic Metadata
    title: str = Field(...)
    """The full title of the paper, including any subtitles if applicable."""

    authors: List[str] = Field(default_factory=list)
    """A list of the paper's authors, typically in the order of contribution."""

    keywords: List[str] = Field(default_factory=list)
    """A list of keywords that summarize the paper's focus and facilitate indexing."""

    publication_year: int = Field(None)
    """The year in which the paper was published."""

    # Core Content Elements
    domain: List[str] = Field(default_factory=list)
    """The research domains or fields addressed by the paper (e.g., ['Natural Language Processing', 'Computer Vision'])."""

    abstract: str = Field(...)
    """A structured abstract that outlines the research problem, methodology, and conclusions in three distinct sections."""

    core_contributions: List[str] = Field(default_factory=list)
    """Key academic contributions that distinguish the paper from prior work in the field."""

    technical_novelty: List[str] = Field(default_factory=list)
    """Specific technical innovations introduced by the research, listed as individual points."""

    # Academic Discussion Dimensions
    research_problem: str = Field("")
    """A clearly defined research question or problem addressed by the study."""

    limitations: List[str] = Field(default_factory=list)
    """An analysis of the methodological or experimental limitations of the research."""

    future_work: List[str] = Field(default_factory=list)
    """Suggestions for potential directions or topics for follow-up studies."""

    impact_analysis: str = Field("")
    """An assessment of the paper's potential influence on the development of the field."""

    def _prepare_vectorization_inner(self) -> str:
        return self.model_dump_json()


class ArticleProposal(ProposedAble, Display):
    """Structured representation of the proposal for an academic paper."""

    title: str = Field(...)
    """The proposed title of the paper."""

    focused_problem: List[str] = Field(default_factory=list)
    """The specific research problem or question that the paper aims to address."""
    research_aim: List[str] = Field(default_factory=list)
    """The main objective or goal of the research, outlining what the study aims to achieve."""
    research_methods: List[str] = Field(default_factory=list)
    """The methods used in the research, including the approach, techniques, and tools employed."""


class ArticleSubsectionOutline(Base):
    """Structured representation of the subsections of an academic paper."""

    title: str = Field(...)
    """The title of the subsection."""

    description: str = Field(...)
    """A brief description of the subsection's content should be, how it fits into the overall structure of the paper, and its significance in the context of the research."""


class ArticleSectionOutline(Base):
    """Structured representation of the sections of an academic paper."""

    title: str = Field(...)
    """The title of the section."""
    description: str = Field(...)
    """A brief description of the section's content should be, how it fits into the overall structure of the paper, and its significance in the context of the research."""
    subsections: List[ArticleSubsectionOutline] = Field(default_factory=list)
    """The subsections of the section, outlining their content and significance."""


class ArticleChapterOutline(Base):
    """Structured representation of the chapters of an academic paper."""

    title: str = Field(...)
    """The title of the chapter."""
    description: str = Field(...)
    """A brief description of the chapter's content should be, how it fits into the overall structure of the paper, and its significance in the context of the research."""
    sections: List[ArticleSectionOutline] = Field(default_factory=list)
    """The sections of the chapter, outlining their content and significance."""


class ArticleOutline(ProposedAble, Display, FinalizedDumpAble):
    """Structured representation of the outline for an academic paper."""

    title: str = Field(...)
    """The proposed title of the paper."""

    prospect: str = Field(...)
    """A brief description of the research problem or question that the paper aims to address manipulating methods or techniques"""

    chapters: List[ArticleChapterOutline] = Field(default_factory=list)
    """The chapters of the paper, outlining their content and significance."""

    def finalized_dump(self) -> str:
        """Finalized dump of the article outline.

        Returns:
            str: The finalized dump of the article outline.
        """
        lines: List[str] = []

        for chapter in self.chapters:
            lines.append(f"= {chapter.title}")
            for section in chapter.sections:
                lines.append(f"== {section.title}")
                for subsection in section.subsections:
                    lines.append(f"=== {subsection.title}")

        return "\n\n".join(lines)



"""This module contains the types for the keyword arguments of the methods in the models module."""

from typing import List, NotRequired, TypedDict

from pydantic import NonNegativeFloat, NonNegativeInt, PositiveInt


class CollectionSimpleConfigKwargs(TypedDict):
    """A type representing the configuration for a collection."""

    dimension: NotRequired[int]
    timeout: NotRequired[float]


class FetchKwargs(TypedDict):
    """A type representing the keyword arguments for the fetch method."""

    collection_name: NotRequired[str]
    similarity_threshold: NotRequired[float]
    result_per_query: NotRequired[int]


class EmbeddingKwargs(TypedDict):
    """A type representing the keyword arguments for the embedding method."""

    model: NotRequired[str]
    dimensions: NotRequired[int]
    timeout: NotRequired[PositiveInt]
    caching: NotRequired[bool]


class LLMKwargs(TypedDict):
    """A type representing the keyword arguments for the LLM (Large Language Model) usage."""

    model: NotRequired[str]
    temperature: NotRequired[NonNegativeFloat]
    stop: NotRequired[str | List[str]]
    top_p: NotRequired[NonNegativeFloat]
    max_tokens: NotRequired[PositiveInt]
    stream: NotRequired[bool]
    timeout: NotRequired[PositiveInt]
    max_retries: NotRequired[PositiveInt]


class ValidateKwargs(LLMKwargs):
    """A type representing the keyword arguments for the validate method."""

    max_validations: NotRequired[PositiveInt]


class GenerateKwargs(ValidateKwargs):
    """A type representing the keyword arguments for the generate method."""

    system_message: NotRequired[str]


class ChooseKwargs(GenerateKwargs):
    """A type representing the keyword arguments for the choose method."""

    k: NotRequired[NonNegativeInt]



"""A module to parse text using regular expressions."""

from typing import Any, Callable, Optional, Self, Tuple, Type

import orjson
import regex
from pydantic import BaseModel, ConfigDict, Field, PositiveInt, PrivateAttr, ValidationError
from regex import Pattern, compile

from fabricatio.journal import logger


class Capture(BaseModel):
    """A class to capture patterns in text using regular expressions.

    Attributes:
        pattern (str): The regular expression pattern to search for.
        _compiled (Pattern): The compiled regular expression pattern.
    """

    model_config = ConfigDict(use_attribute_docstrings=True)
    target_groups: Tuple[int, ...] = Field(default_factory=tuple)
    """The target groups to capture from the pattern."""
    pattern: str = Field(frozen=True)
    """The regular expression pattern to search for."""
    flags: PositiveInt = Field(default=regex.DOTALL | regex.MULTILINE | regex.IGNORECASE, frozen=True)
    """The flags to use when compiling the regular expression pattern."""
    _compiled: Pattern = PrivateAttr()

    def model_post_init(self, __context: Any) -> None:
        """Initialize the compiled pattern."""
        self._compiled = compile(self.pattern, self.flags)

    def capture(self, text: str) -> Tuple[str, ...] | str | None:
        """Capture the first occurrence of the pattern in the given text.

        Args:
            text (str): The text to search the pattern in.

        Returns:
            str | None: The captured text if the pattern is found, otherwise None.

        """
        match = self._compiled.search(text)
        if match is None:
            return None

        if self.target_groups:
            cap = tuple(match.group(g) for g in self.target_groups)
            logger.debug(f"Captured text: {'\n\n'.join(cap)}")
            return cap
        cap = match.group(1)
        logger.debug(f"Captured text: \n{cap}")
        return cap

    def convert_with[T](self, text: str, convertor: Callable[[Tuple[str, ...]], T] | Callable[[str], T]) -> T | None:
        """Convert the given text using the pattern.

        Args:
            text (str): The text to search the pattern in.
            convertor (Callable[[Tuple[str, ...]], T] | Callable[[str], T]): The function to convert the captured text.

        Returns:
            str | None: The converted text if the pattern is found, otherwise None.
        """
        if (cap := self.capture(text)) is None:
            return None
        try:
            return convertor(cap)
        except (ValueError, SyntaxError, ValidationError) as e:
            logger.error(f"Failed to convert text using {convertor.__name__} to convert.\nerror: {e}\n {cap}")
            return None

    def validate_with[K, T, E](
        self,
        text: str,
        target_type: Type[T],
        elements_type: Optional[Type[E]] = None,
        length: Optional[int] = None,
        deserializer: Callable[[Tuple[str, ...]], K] | Callable[[str], K] = orjson.loads,
    ) -> T | None:
        """Validate the given text using the pattern.

        Args:
            text (str): The text to search the pattern in.
            target_type (Type[T]): The expected type of the output, dict or list.
            elements_type (Optional[Type[E]]): The expected type of the elements in the output dict keys or list elements.
            length (Optional[int]): The expected length of the output, bool(length)==False means no length validation.
            deserializer (Callable[[Tuple[str, ...]], K] | Callable[[str], K]): The function to deserialize the captured text.

        Returns:
            T | None: The validated text if the pattern is found and the output is of the expected type, otherwise None.
        """
        judges = [lambda output_obj: isinstance(output_obj, target_type)]
        if elements_type:
            judges.append(lambda output_obj: all(isinstance(e, elements_type) for e in output_obj))
        if length:
            judges.append(lambda output_obj: len(output_obj) == length)

        if (out := self.convert_with(text, deserializer)) and all(j(out) for j in judges):
            return out
        return None

    @classmethod
    def capture_code_block(cls, language: str) -> Self:
        """Capture the first occurrence of a code block in the given text.

        Args:
            language (str): The text containing the code block.

        Returns:
            Self: The instance of the class with the captured code block.
        """
        return cls(pattern=f"```{language}\n(.*?)\n```")


JsonCapture = Capture.capture_code_block("json")
PythonCapture = Capture.capture_code_block("python")
MarkdownCapture = Capture.capture_code_block("markdown")
CodeBlockCapture = Capture(pattern="```.*?\n(.*?)\n```")



"""Logging setup for the project."""

import sys

from loguru import logger
from rich import pretty, traceback

from fabricatio.config import configs

pretty.install()
traceback.install()
logger.remove()
logger.add(
    configs.debug.log_file,
    level=configs.debug.log_level,
    rotation=f"{configs.debug.rotation} weeks",
    retention=f"{configs.debug.retention} weeks",
)
logger.add(sys.stderr, level=configs.debug.log_level)


if __name__ == "__main__":
    logger.debug("This is a trace message.")
    logger.info("This is an information message.")
    logger.success("This is a success message.")
    logger.warning("This is a warning message.")
    logger.error("This is an error message.")
    logger.critical("This is a critical message.")



"""A module for the task capabilities of the Fabricatio library."""

from typing import List, Type, Unpack, overload

from fabricatio.models.generic import ProposedAble
from fabricatio.models.kwargs_types import GenerateKwargs
from fabricatio.models.usages import LLMUsage


class Propose[M: ProposedAble](LLMUsage):
    """A class that proposes an Obj based on a prompt."""

    @overload
    async def propose(
        self,
        cls: Type[M],
        prompt: List[str],
        **kwargs: Unpack[GenerateKwargs],
    ) -> List[M]: ...

    @overload
    async def propose(
        self,
        cls: Type[M],
        prompt: str,
        **kwargs: Unpack[GenerateKwargs],
    ) -> M: ...

    async def propose(
        self,
        cls: Type[M],
        prompt: List[str] | str,
        **kwargs: Unpack[GenerateKwargs],
    ) -> List[M] | M:
        """Asynchronously proposes a task based on a given prompt and parameters.

        Parameters:
            cls: The class type of the task to be proposed.
            prompt: The prompt text for proposing a task, which is a string that must be provided.
            **kwargs: The keyword arguments for the LLM (Large Language Model) usage.

        Returns:
            A Task object based on the proposal result.
        """
        if isinstance(prompt, str):
            return await self.aask_validate(
                question=cls.create_json_prompt(prompt),
                validator=cls.instantiate_from_string,
                **kwargs,
            )
        return await self.aask_validate_batch(
            questions=[cls.create_json_prompt(p) for p in prompt],
            validator=cls.instantiate_from_string,
            **kwargs,
        )



"""This module defines generic classes for models in the Fabricatio library."""

from abc import abstractmethod
from pathlib import Path
from typing import Callable, Iterable, List, Optional, Self, Union, final

import orjson
from fabricatio._rust import blake3_hash
from fabricatio._rust_instances import template_manager
from fabricatio.config import configs
from fabricatio.fs.readers import magika, safe_text_read
from fabricatio.journal import logger
from fabricatio.parser import JsonCapture
from pydantic import (
    BaseModel,
    ConfigDict,
    Field,
    HttpUrl,
    NonNegativeFloat,
    PositiveFloat,
    PositiveInt,
    SecretStr,
)


class Base(BaseModel):
    """Base class for all models with Pydantic configuration."""

    model_config = ConfigDict(use_attribute_docstrings=True)


class Display(Base):
    """Class that provides a method to display the model in a formatted JSON string."""

    def display(self) -> str:
        """Display the model in a formatted JSON string.

        Returns:
            str: The formatted JSON string of the model.
        """
        return self.model_dump_json(indent=1)


class Named(Base):
    """Class that includes a name attribute."""

    name: str = Field(frozen=True)
    """The name of the object."""


class Described(Base):
    """Class that includes a description attribute."""

    description: str = Field(default="", frozen=True)
    """The description of the object."""


class WithBriefing(Named, Described):
    """Class that provides a briefing based on the name and description."""

    @property
    def briefing(self) -> str:
        """Get the briefing of the object.

        Returns:
            str: The briefing of the object.
        """
        return f"{self.name}: {self.description}" if self.description else self.name


class WithFormatedJsonSchema(Base):
    """Class that provides a formatted JSON schema of the model."""

    @classmethod
    def formated_json_schema(cls) -> str:
        """Get the JSON schema of the model in a formatted string.

        Returns:
            str: The JSON schema of the model in a formatted string.
        """
        return orjson.dumps(
            cls.model_json_schema(),
            option=orjson.OPT_INDENT_2 | orjson.OPT_SORT_KEYS,
        ).decode()


class CreateJsonObjPrompt(WithFormatedJsonSchema):
    """Class that provides a prompt for creating a JSON object."""

    @classmethod
    def create_json_prompt(cls, requirement: str) -> str:
        """Create the prompt for creating a JSON object with given requirement.

        Args:
            requirement (str): The requirement for the JSON object.

        Returns:
            str: The prompt for creating a JSON object with given requirement.
        """
        return template_manager.render_template(
            configs.templates.create_json_obj_template,
            {"requirement": requirement, "json_schema": cls.formated_json_schema()},
        )


class InstantiateFromString(Base):
    """Class that provides a method to instantiate the class from a string."""

    @classmethod
    def instantiate_from_string(cls, string: str) -> Self | None:
        """Instantiate the class from a string.

        Args:
            string (str): The string to instantiate the class from.

        Returns:
            Self | None: The instance of the class or None if the string is not valid.
        """
        return JsonCapture.convert_with(string, cls.model_validate_json)


class ProposedAble(CreateJsonObjPrompt, InstantiateFromString):
    """Class that provides methods for proposing a task."""

    pass


class FinalizedDumpAble(Base):
    """Class that provides a method to finalize the dump of the object."""

    @abstractmethod
    def finalized_dump(self) -> str:
        """Finalize the dump of the object.

        Returns:
            str: The finalized dump of the object.
        """

    def finalized_dump_to(self, path: str | Path) -> Self:
        """Finalize the dump of the object to a file.

        Args:
            path (str | Path): The path to save the finalized dump.

        Returns:
            Self: The current instance of the object.
        """
        Path(path).write_text(self.finalized_dump(), encoding="utf-8")
        return self


class WithDependency(Base):
    """Class that manages file dependencies."""

    dependencies: List[str] = Field(default_factory=list)
    """The file dependencies which is needed to read or write to meet a specific requirement, a list of file paths."""

    def add_dependency[P: str | Path](self, dependency: P | List[P]) -> Self:
        """Add a file dependency to the task.

        Args:
            dependency (str | Path | List[str | Path]): The file dependency to add to the task.

        Returns:
            Self: The current instance of the task.
        """
        if not isinstance(dependency, list):
            dependency = [dependency]
        self.dependencies.extend(Path(d).as_posix() for d in dependency)
        return self

    def remove_dependency[P: str | Path](self, dependency: P | List[P]) -> Self:
        """Remove a file dependency from the task.

        Args:
            dependency (str | Path | List[str | Path]): The file dependency to remove from the task.

        Returns:
            Self: The current instance of the task.
        """
        if not isinstance(dependency, list):
            dependency = [dependency]
        for d in dependency:
            self.dependencies.remove(Path(d).as_posix())
        return self

    def clear_dependencies(self) -> Self:
        """Clear all file dependencies from the task.

        Returns:
            Self: The current instance of the task.
        """
        self.dependencies.clear()
        return self

    def override_dependencies[P: str | Path](self, dependencies: List[P] | P) -> Self:
        """Override the file dependencies of the task.

        Args:
            dependencies (List[str | Path] | str | Path): The file dependencies to override the task's dependencies.

        Returns:
            Self: The current instance of the task.
        """
        return self.clear_dependencies().add_dependency(dependencies)

    def pop_dependence[T](self, idx: int = -1, reader: Callable[[str], T] = safe_text_read) -> T:
        """Pop the file dependencies from the task.

        Returns:
            str: The popped file dependency
        """
        return reader(self.dependencies.pop(idx))

    @property
    def dependencies_prompt(self) -> str:
        """Generate a prompt for the task based on the file dependencies.

        Returns:
            str: The generated prompt for the task.
        """
        return template_manager.render_template(
            configs.templates.dependencies_template,
            {
                (pth := Path(p)).name: {
                    "path": pth.as_posix(),
                    "exists": pth.exists(),
                    "description": (identity := magika.identify_path(pth)).output.description,
                    "size": f"{pth.stat().st_size / (1024 * 1024) if pth.exists() and pth.is_file() else 0:.3f} MB",
                    "content": (text := safe_text_read(pth)),
                    "lines": len(text.splitlines()),
                    "language": identity.output.ct_label,
                    "checksum": blake3_hash(pth.read_bytes()) if pth.exists() and pth.is_file() else "unknown",
                }
                for p in self.dependencies
            },
        )


class PrepareVectorization(Base):
    """Class that prepares the vectorization of the model."""

    @abstractmethod
    def _prepare_vectorization_inner(self) -> str:
        """Prepare the vectorization of the model."""

    def prepare_vectorization(self, max_length: Optional[int] = None) -> str:
        """Prepare the vectorization of the model.

        Returns:
            str: The prepared vectorization of the model.
        """
        max_length = max_length or configs.embedding.max_sequence_length
        chunk = self._prepare_vectorization_inner()
        if len(chunk) > max_length:
            logger.error(err := f"Chunk exceeds maximum sequence length {max_length}.")
            raise ValueError(err)

        return chunk


class ScopedConfig(Base):
    """Class that manages a scoped configuration."""

    llm_api_endpoint: Optional[HttpUrl] = None
    """The OpenAI API endpoint."""

    llm_api_key: Optional[SecretStr] = None
    """The OpenAI API key."""

    llm_timeout: Optional[PositiveInt] = None
    """The timeout of the LLM model."""

    llm_max_retries: Optional[PositiveInt] = None
    """The maximum number of retries."""

    llm_model: Optional[str] = None
    """The LLM model name."""

    llm_temperature: Optional[NonNegativeFloat] = None
    """The temperature of the LLM model."""

    llm_stop_sign: Optional[str | List[str]] = None
    """The stop sign of the LLM model."""

    llm_top_p: Optional[NonNegativeFloat] = None
    """The top p of the LLM model."""

    llm_generation_count: Optional[PositiveInt] = None
    """The number of generations to generate."""

    llm_stream: Optional[bool] = None
    """Whether to stream the LLM model's response."""

    llm_max_tokens: Optional[PositiveInt] = None
    """The maximum number of tokens to generate."""

    embedding_api_endpoint: Optional[HttpUrl] = None
    """The OpenAI API endpoint."""

    embedding_api_key: Optional[SecretStr] = None
    """The OpenAI API key."""

    embedding_timeout: Optional[PositiveInt] = None
    """The timeout of the LLM model."""

    embedding_model: Optional[str] = None
    """The LLM model name."""

    embedding_max_sequence_length: Optional[PositiveInt] = None
    """The maximum sequence length."""

    embedding_dimensions: Optional[PositiveInt] = None
    """The dimensions of the embedding."""
    embedding_caching: Optional[bool] = False
    """Whether to cache the embedding result."""

    milvus_uri: Optional[HttpUrl] = Field(default=None)
    """The URI of the Milvus server."""
    milvus_token: Optional[SecretStr] = Field(default=None)
    """The token for the Milvus server."""
    milvus_timeout: Optional[PositiveFloat] = Field(default=None)
    """The timeout for the Milvus server."""
    milvus_dimensions: Optional[PositiveInt] = Field(default=None)
    """The dimensions of the Milvus server."""

    @final
    def fallback_to(self, other: "ScopedConfig") -> Self:
        """Fallback to another instance's attribute values if the current instance's attributes are None.

        Args:
            other (LLMUsage): Another instance from which to copy attribute values.

        Returns:
            Self: The current instance, allowing for method chaining.
        """
        # Iterate over the attribute names and copy values from 'other' to 'self' where applicable
        # noinspection PydanticTypeChecker,PyTypeChecker
        for attr_name in ScopedConfig.model_fields:
            # Copy the attribute value from 'other' to 'self' only if 'self' has None and 'other' has a non-None value
            if getattr(self, attr_name) is None and (attr := getattr(other, attr_name)) is not None:
                setattr(self, attr_name, attr)

        # Return the current instance to allow for method chaining
        return self

    @final
    def hold_to(self, others: Union["ScopedConfig", Iterable["ScopedConfig"]]) -> Self:
        """Hold to another instance's attribute values if the current instance's attributes are None.

        Args:
            others (LLMUsage | Iterable[LLMUsage]): Another instance or iterable of instances from which to copy attribute values.

        Returns:
            Self: The current instance, allowing for method chaining.
        """
        if not isinstance(others, Iterable):
            others = [others]
        for other in others:
            # noinspection PyTypeChecker,PydanticTypeChecker
            for attr_name in ScopedConfig.model_fields:
                if (attr := getattr(self, attr_name)) is not None and getattr(other, attr_name) is None:
                    setattr(other, attr_name, attr)



"""Actions for transmitting tasks to targets."""

from os import PathLike
from pathlib import Path
from typing import Callable, List

from fabricatio.fs import safe_text_read
from fabricatio.journal import logger
from fabricatio.models.action import Action
from fabricatio.models.extra import ArticleEssence, ArticleOutline, ArticleProposal
from fabricatio.models.task import Task


class ExtractArticleEssence(Action):
    """Extract the essence of article(s) in text format from the paths specified in the task dependencies.

    Notes:
        This action is designed to extract vital information from articles with Markdown format, which is pure text, and
        which is converted from pdf files using `magic-pdf` from the `MinerU` project, see https://github.com/opendatalab/MinerU
    """

    output_key: str = "article_essence"
    """The key of the output data."""

    async def _execute[P: PathLike | str](
        self,
        task_input: Task,
        reader: Callable[[P], str] = lambda p: Path(p).read_text(encoding="utf-8"),
        **_,
    ) -> List[ArticleEssence]:
        if not task_input.dependencies:
            logger.info(err := "Task not approved, since no dependencies are provided.")
            raise RuntimeError(err)

        # trim the references
        contents = ["References".join(c.split("References")[:-1]) for c in map(reader, task_input.dependencies)]
        return await self.propose(
            ArticleEssence,
            contents,
            system_message=f"# your personal briefing: \n{self.briefing}",
        )


class GenerateArticleProposal(Action):
    """Generate an outline for the article based on the extracted essence."""

    output_key: str = "article_proposal"
    """The key of the output data."""

    async def _execute(
        self,
        task_input: Task,
        **_,
    ) -> ArticleProposal:
        input_path = await self.awhich_pathstr(
            f"{task_input.briefing}\nExtract the path of file, which contains the article briefing that I need to read."
        )

        return await self.propose(
            ArticleProposal,
            safe_text_read(input_path),
            system_message=f"# your personal briefing: \n{self.briefing}",
        )


class GenerateOutline(Action):
    """Generate the article based on the outline."""

    output_key: str = "article"
    """The key of the output data."""

    async def _execute(
        self,
        article_proposal: ArticleProposal,
        **_,
    ) -> ArticleOutline:
        return await self.propose(
            ArticleOutline,
            article_proposal.display(),
            system_message=f"# your personal briefing: \n{self.briefing}",
        )



"""Fabricatio is a Python library for building llm app using event-based agent structure."""

from importlib.util import find_spec

from fabricatio._rust_instances import template_manager
from fabricatio.actions.article import ExtractArticleEssence, GenerateArticleProposal, GenerateOutline
from fabricatio.actions.output import DumpFinalizedOutput
from fabricatio.core import env
from fabricatio.fs import magika, safe_json_read, safe_text_read
from fabricatio.journal import logger
from fabricatio.models.action import Action, WorkFlow
from fabricatio.models.events import Event
from fabricatio.models.extra import ArticleEssence
from fabricatio.models.role import Role
from fabricatio.models.task import Task
from fabricatio.models.tool import ToolBox
from fabricatio.models.utils import Message, Messages
from fabricatio.parser import Capture, CodeBlockCapture, JsonCapture, PythonCapture
from fabricatio.toolboxes import arithmetic_toolbox, basic_toolboxes, fs_toolbox
from fabricatio.workflows.articles import WriteOutlineWorkFlow

__all__ = [
    "Action",
    "ArticleEssence",
    "Capture",
    "CodeBlockCapture",
    "DumpFinalizedOutput",
    "Event",
    "ExtractArticleEssence",
    "GenerateArticleProposal",
    "GenerateOutline",
    "JsonCapture",
    "Message",
    "Messages",
    "PythonCapture",
    "Role",
    "Task",
    "ToolBox",
    "WorkFlow",
    "WriteOutlineWorkFlow",
    "arithmetic_toolbox",
    "basic_toolboxes",
    "env",
    "fs_toolbox",
    "logger",
    "magika",
    "safe_json_read",
    "safe_text_read",
    "template_manager",
]


if find_spec("pymilvus"):
    from fabricatio.actions.rag import InjectToDB
    from fabricatio.capabilities.rag import RAG
    from fabricatio.workflows.rag import StoreArticle

    __all__ += ["RAG", "InjectToDB", "StoreArticle"]



"""Dump the finalized output to a file."""

from typing import Unpack

from fabricatio.models.action import Action
from fabricatio.models.generic import FinalizedDumpAble
from fabricatio.models.task import Task


class DumpFinalizedOutput(Action):
    """Dump the finalized output to a file."""

    output_key: str = "dump_path"

    async def _execute(self, task_input: Task, to_dump: FinalizedDumpAble, **cxt: Unpack) -> str:
        dump_path = await self.awhich_pathstr(
            f"{task_input.briefing}\n\nExtract a single path of the file, to which I will dump the data."
        )

        to_dump.finalized_dump_to(dump_path)
        return dump_path



"""This module contains classes that manage the usage of language models and tools in tasks."""

from asyncio import gather
from typing import Callable, Dict, Iterable, List, Optional, Self, Set, Type, Union, Unpack, overload

import asyncstdlib
import litellm
from fabricatio._rust_instances import template_manager
from fabricatio.config import configs
from fabricatio.journal import logger
from fabricatio.models.generic import ScopedConfig, WithBriefing
from fabricatio.models.kwargs_types import ChooseKwargs, EmbeddingKwargs, GenerateKwargs, LLMKwargs
from fabricatio.models.task import Task
from fabricatio.models.tool import Tool, ToolBox
from fabricatio.models.utils import Messages
from fabricatio.parser import JsonCapture
from litellm import stream_chunk_builder
from litellm.types.utils import (
    Choices,
    EmbeddingResponse,
    ModelResponse,
    StreamingChoices,
)
from litellm.utils import CustomStreamWrapper
from more_itertools import duplicates_everseen
from pydantic import Field, NonNegativeInt, PositiveInt


class LLMUsage(ScopedConfig):
    """Class that manages LLM (Large Language Model) usage parameters and methods."""

    @classmethod
    def _scoped_model(cls) -> Type["LLMUsage"]:
        return LLMUsage

    async def aquery(
        self,
        messages: List[Dict[str, str]],
        n: PositiveInt | None = None,
        **kwargs: Unpack[LLMKwargs],
    ) -> ModelResponse | CustomStreamWrapper:
        """Asynchronously queries the language model to generate a response based on the provided messages and parameters.

        Args:
            messages (List[Dict[str, str]]): A list of messages, where each message is a dictionary containing the role and content of the message.
            n (PositiveInt | None): The number of responses to generate. Defaults to the instance's `llm_generation_count` or the global configuration.
            **kwargs (Unpack[LLMKwargs]): Additional keyword arguments for the LLM usage.

        Returns:
            ModelResponse | CustomStreamWrapper: An object containing the generated response and other metadata from the model.
        """
        # Call the underlying asynchronous completion function with the provided and default parameters
        return await litellm.acompletion(
            messages=messages,
            n=n or self.llm_generation_count or configs.llm.generation_count,
            model=kwargs.get("model") or self.llm_model or configs.llm.model,
            temperature=kwargs.get("temperature") or self.llm_temperature or configs.llm.temperature,
            stop=kwargs.get("stop") or self.llm_stop_sign or configs.llm.stop_sign,
            top_p=kwargs.get("top_p") or self.llm_top_p or configs.llm.top_p,
            max_tokens=kwargs.get("max_tokens") or self.llm_max_tokens or configs.llm.max_tokens,
            stream=kwargs.get("stream") or self.llm_stream or configs.llm.stream,
            timeout=kwargs.get("timeout") or self.llm_timeout or configs.llm.timeout,
            max_retries=kwargs.get("max_retries") or self.llm_max_retries or configs.llm.max_retries,
            api_key=(self.llm_api_key or configs.llm.api_key).get_secret_value(),
            base_url=(self.llm_api_endpoint or configs.llm.api_endpoint).unicode_string(),
        )

    async def ainvoke(
        self,
        question: str,
        system_message: str = "",
        n: PositiveInt | None = None,
        **kwargs: Unpack[LLMKwargs],
    ) -> List[Choices | StreamingChoices]:
        """Asynchronously invokes the language model with a question and optional system message.

        Args:
            question (str): The question to ask the model.
            system_message (str): The system message to provide context to the model. Defaults to an empty string.
            n (PositiveInt | None): The number of responses to generate. Defaults to the instance's `llm_generation_count` or the global configuration.
            **kwargs (Unpack[LLMKwargs]): Additional keyword arguments for the LLM usage.

        Returns:
            List[Choices | StreamingChoices]: A list of choices or streaming choices from the model response.
        """
        resp = await self.aquery(
            messages=Messages().add_system_message(system_message).add_user_message(question),
            n=n,
            **kwargs,
        )
        if isinstance(resp, ModelResponse):
            return resp.choices
        if isinstance(resp, CustomStreamWrapper):
            if not configs.debug.streaming_visible:
                return stream_chunk_builder(await asyncstdlib.list()).choices
            chunks = []
            async for chunk in resp:
                chunks.append(chunk)
                print(chunk.choices[0].delta.content or "", end="")  # noqa: T201
            return stream_chunk_builder(chunks).choices
        logger.critical(err := f"Unexpected response type: {type(resp)}")
        raise ValueError(err)

    @overload
    async def aask(
        self,
        question: List[str],
        system_message: Optional[List[str]] = None,
        **kwargs: Unpack[LLMKwargs],
    ) -> List[str]: ...
    @overload
    async def aask(
        self,
        question: str,
        system_message: Optional[List[str]] = None,
        **kwargs: Unpack[LLMKwargs],
    ) -> List[str]: ...
    @overload
    async def aask(
        self,
        question: List[str],
        system_message: Optional[str] = None,
        **kwargs: Unpack[LLMKwargs],
    ) -> List[str]: ...

    @overload
    async def aask(
        self,
        question: str,
        system_message: Optional[str] = None,
        **kwargs: Unpack[LLMKwargs],
    ) -> str: ...

    async def aask(
        self,
        question: str | List[str],
        system_message: Optional[str | List[str]] = None,
        **kwargs: Unpack[LLMKwargs],
    ) -> str | List[str]:
        """Asynchronously asks the language model a question and returns the response content.

        Args:
            question (str | List[str]): The question to ask the model.
            system_message (str | List[str] | None): The system message to provide context to the model. Defaults to an empty string.
            **kwargs (Unpack[LLMKwargs]): Additional keyword arguments for the LLM usage.

        Returns:
            str | List[str]: The content of the model's response message.
        """
        system_message = system_message or ""
        match (isinstance(question, list), isinstance(system_message, list)):
            case (True, True):
                res = await gather(
                    *[
                        self.ainvoke(n=1, question=q, system_message=sm, **kwargs)
                        for q, sm in zip(question, system_message, strict=True)
                    ]
                )
                return [r.pop().message.content for r in res]
            case (True, False):
                res = await gather(
                    *[self.ainvoke(n=1, question=q, system_message=system_message, **kwargs) for q in question]
                )
                return [r.pop().message.content for r in res]
            case (False, True):
                res = await gather(
                    *[self.ainvoke(n=1, question=question, system_message=sm, **kwargs) for sm in system_message]
                )
                return [r.pop().message.content for r in res]
            case (False, False):
                return (
                    (
                        await self.ainvoke(
                            n=1,
                            question=question,
                            system_message=system_message,
                            **kwargs,
                        )
                    ).pop()
                ).message.content
            case _:
                raise RuntimeError("Should not reach here.")

    async def aask_validate[T](
        self,
        question: str,
        validator: Callable[[str], T | None],
        max_validations: PositiveInt = 2,
        system_message: str = "",
        **kwargs: Unpack[LLMKwargs],
    ) -> T:
        """Asynchronously asks a question and validates the response using a given validator.

        Args:
            question (str): The question to ask.
            validator (Callable[[str], T | None]): A function to validate the response.
            max_validations (PositiveInt): Maximum number of validation attempts. Defaults to 2.
            system_message (str): System message to include in the request. Defaults to an empty string.
            **kwargs (Unpack[LLMKwargs]): Additional keyword arguments for the LLM usage.

        Returns:
            T: The validated response.

        Raises:
            ValueError: If the response fails to validate after the maximum number of attempts.
        """
        for i in range(max_validations):
            if (
                response := await self.aask(
                    question=question,
                    system_message=system_message,
                    **kwargs,
                )
            ) and (validated := validator(response)):
                logger.debug(f"Successfully validated the response at {i}th attempt.")
                return validated
        logger.error(err := f"Failed to validate the response after {max_validations} attempts.")
        raise ValueError(err)

    async def aask_validate_batch[T](
        self,
        questions: List[str],
        validator: Callable[[str], T | None],
        **kwargs: Unpack[GenerateKwargs],
    ) -> List[T]:
        """Asynchronously asks a batch of questions and validates the responses using a given validator.

        Args:
            questions (List[str]): The list of questions to ask.
            validator (Callable[[str], T | None]): A function to validate the response.
            **kwargs (Unpack[GenerateKwargs]): Additional keyword arguments for the LLM usage.

        Returns:
            T: The validated response.

        Raises:
            ValueError: If the response fails to validate after the maximum number of attempts.
        """
        return await gather(*[self.aask_validate(question, validator, **kwargs) for question in questions])

    async def aliststr(self, requirement: str, k: NonNegativeInt = 0, **kwargs: Unpack[GenerateKwargs]) -> List[str]:
        """Asynchronously generates a list of strings based on a given requirement.

        Args:
            requirement (str): The requirement for the list of strings.
            k (NonNegativeInt): The number of choices to select, 0 means infinite. Defaults to 0.
            **kwargs (Unpack[GenerateKwargs]): Additional keyword arguments for the LLM usage.

        Returns:
            List[str]: The validated response as a list of strings.
        """
        return await self.aask_validate(
            template_manager.render_template(
                configs.templates.liststr_template,
                {"requirement": requirement, "k": k},
            ),
            lambda resp: JsonCapture.validate_with(resp, target_type=list, elements_type=str, length=k),
            **kwargs,
        )

    async def apathstr(self, requirement: str, **kwargs: Unpack[ChooseKwargs]) -> List[str]:
        """Asynchronously generates a list of strings based on a given requirement.

        Args:
            requirement (str): The requirement for the list of strings.
            **kwargs (Unpack[ChooseKwargs]): Additional keyword arguments for the LLM usage.

        Returns:
            List[str]: The validated response as a list of strings.
        """
        return await self.aliststr(
            template_manager.render_template(
                configs.templates.pathstr_template,
                {"requirement": requirement},
            ),
            **kwargs,
        )

    async def awhich_pathstr(self, requirement: str, **kwargs: Unpack[GenerateKwargs]) -> str:
        """Asynchronously generates a single path string based on a given requirement.

        Args:
            requirement (str): The requirement for the list of strings.
            **kwargs (Unpack[GenerateKwargs]): Additional keyword arguments for the LLM usage.

        Returns:
            str: The validated response as a single string.
        """
        return (
            await self.apathstr(
                requirement,
                k=1,
                **kwargs,
            )
        ).pop()

    async def achoose[T: WithBriefing](
        self,
        instruction: str,
        choices: List[T],
        k: NonNegativeInt = 0,
        **kwargs: Unpack[GenerateKwargs],
    ) -> List[T]:
        """Asynchronously executes a multi-choice decision-making process, generating a prompt based on the instruction and options, and validates the returned selection results.

        Args:
            instruction (str): The user-provided instruction/question description.
            choices (List[T]): A list of candidate options, requiring elements to have `name` and `briefing` fields.
            k (NonNegativeInt): The number of choices to select, 0 means infinite. Defaults to 0.
            **kwargs (Unpack[GenerateKwargs]): Additional keyword arguments for the LLM usage.

        Returns:
            List[T]: The final validated selection result list, with element types matching the input `choices`.

        Important:
            - Uses a template engine to generate structured prompts.
            - Ensures response compliance through JSON parsing and format validation.
            - Relies on `aask_validate` to implement retry mechanisms with validation.
        """
        if dup := duplicates_everseen(choices, key=lambda x: x.name):
            logger.error(err := f"Redundant choices: {dup}")
            raise ValueError(err)
        prompt = template_manager.render_template(
            configs.templates.make_choice_template,
            {
                "instruction": instruction,
                "options": [m.model_dump(include={"name", "briefing"}) for m in choices],
                "k": k,
            },
        )
        names = {c.name for c in choices}

        logger.debug(f"Start choosing between {names} with prompt: \n{prompt}")

        def _validate(response: str) -> List[T] | None:
            ret = JsonCapture.validate_with(response, target_type=List, elements_type=str, length=k)
            if ret is None or set(ret) - names:
                return None
            return [
                next(candidate for candidate in choices if candidate.name == candidate_name) for candidate_name in ret
            ]

        return await self.aask_validate(
            question=prompt,
            validator=_validate,
            **kwargs,
        )

    async def apick[T: WithBriefing](
        self,
        instruction: str,
        choices: List[T],
        **kwargs: Unpack[GenerateKwargs],
    ) -> T:
        """Asynchronously picks a single choice from a list of options using AI validation.

        Args:
            instruction (str): The user-provided instruction/question description.
            choices (List[T]): A list of candidate options, requiring elements to have `name` and `briefing` fields.
            **kwargs (Unpack[GenerateKwargs]): Additional keyword arguments for the LLM usage.

        Returns:
            T: The single selected item from the choices list.

        Raises:
            ValueError: If validation fails after maximum attempts or if no valid selection is made.
        """
        return (
            await self.achoose(
                instruction=instruction,
                choices=choices,
                k=1,
                **kwargs,
            )
        )[0]

    async def ajudge(
        self,
        prompt: str,
        affirm_case: str = "",
        deny_case: str = "",
        **kwargs: Unpack[GenerateKwargs],
    ) -> bool:
        """Asynchronously judges a prompt using AI validation.

        Args:
            prompt (str): The input prompt to be judged.
            affirm_case (str): The affirmative case for the AI model. Defaults to an empty string.
            deny_case (str): The negative case for the AI model. Defaults to an empty string.
            **kwargs (Unpack[GenerateKwargs]): Additional keyword arguments for the LLM usage.

        Returns:
            bool: The judgment result (True or False) based on the AI's response.
        """
        return await self.aask_validate(
            question=template_manager.render_template(
                configs.templates.make_judgment_template,
                {"prompt": prompt, "affirm_case": affirm_case, "deny_case": deny_case},
            ),
            validator=lambda resp: JsonCapture.validate_with(resp, target_type=bool),
            **kwargs,
        )


class EmbeddingUsage(LLMUsage):
    """A class representing the embedding model."""

    async def aembedding(
        self,
        input_text: List[str],
        model: Optional[str] = None,
        dimensions: Optional[int] = None,
        timeout: Optional[PositiveInt] = None,
        caching: Optional[bool] = False,
    ) -> EmbeddingResponse:
        """Asynchronously generates embeddings for the given input text.

        Args:
            input_text (List[str]): A list of strings to generate embeddings for.
            model (Optional[str]): The model to use for embedding. Defaults to the instance's `llm_model` or the global configuration.
            dimensions (Optional[int]): The dimensions of the embedding output should have, which is used to validate the result. Defaults to None.
            timeout (Optional[PositiveInt]): The timeout for the embedding request. Defaults to the instance's `llm_timeout` or the global configuration.
            caching (Optional[bool]): Whether to cache the embedding result. Defaults to False.


        Returns:
            EmbeddingResponse: The response containing the embeddings.
        """
        # check seq length
        max_len = self.embedding_max_sequence_length or configs.embedding.max_sequence_length
        if any(len(t) > max_len for t in input_text):
            logger.error(err := f"Input text exceeds maximum sequence length {max_len}.")
            raise ValueError(err)

        return await litellm.aembedding(
            input=input_text,
            caching=caching or self.embedding_caching or configs.embedding.caching,
            dimensions=dimensions or self.embedding_dimensions or configs.embedding.dimensions,
            model=model or self.embedding_model or configs.embedding.model or self.llm_model or configs.llm.model,
            timeout=timeout
            or self.embedding_timeout
            or configs.embedding.timeout
            or self.llm_timeout
            or configs.llm.timeout,
            api_key=(
                self.embedding_api_key or configs.embedding.api_key or self.llm_api_key or configs.llm.api_key
            ).get_secret_value(),
            api_base=(
                self.embedding_api_endpoint
                or configs.embedding.api_endpoint
                or self.llm_api_endpoint
                or configs.llm.api_endpoint
            )
            .unicode_string()
            .rstrip("/"),
            # seems embedding function takes no base_url end with a slash
        )

    @overload
    async def vectorize(self, input_text: List[str], **kwargs: Unpack[EmbeddingKwargs]) -> List[List[float]]: ...
    @overload
    async def vectorize(self, input_text: str, **kwargs: Unpack[EmbeddingKwargs]) -> List[float]: ...

    async def vectorize(
        self, input_text: List[str] | str, **kwargs: Unpack[EmbeddingKwargs]
    ) -> List[List[float]] | List[float]:
        """Asynchronously generates vector embeddings for the given input text.

        Args:
            input_text (List[str] | str): A string or list of strings to generate embeddings for.
            **kwargs (Unpack[EmbeddingKwargs]): Additional keyword arguments for embedding.

        Returns:
            List[List[float]] | List[float]: The generated embeddings.
        """
        if isinstance(input_text, str):
            return (await self.aembedding([input_text], **kwargs)).data[0].get("embedding")

        return [o.get("embedding") for o in (await self.aembedding(input_text, **kwargs)).data]


class ToolBoxUsage(LLMUsage):
    """A class representing the usage of tools in a task."""

    toolboxes: Set[ToolBox] = Field(default_factory=set)
    """A set of toolboxes used by the instance."""

    @property
    def available_toolbox_names(self) -> List[str]:
        """Return a list of available toolbox names."""
        return [toolbox.name for toolbox in self.toolboxes]

    async def choose_toolboxes(
        self,
        task: Task,
        system_message: str = "",
        k: NonNegativeInt = 0,
        max_validations: PositiveInt = 2,
        **kwargs: Unpack[LLMKwargs],
    ) -> List[ToolBox]:
        """Asynchronously executes a multi-choice decision-making process to choose toolboxes.

        Args:
            task (Task): The task for which to choose toolboxes.
            system_message (str): Custom system-level prompt, defaults to an empty string.
            k (NonNegativeInt): The number of toolboxes to select, 0 means infinite. Defaults to 0.
            max_validations (PositiveInt): Maximum number of validation failures, default is 2.
            **kwargs (Unpack[LLMKwargs]): Additional keyword arguments for the LLM usage.

        Returns:
            List[ToolBox]: The selected toolboxes.
        """
        if not self.toolboxes:
            logger.warning("No toolboxes available.")
            return []
        return await self.achoose(
            instruction=task.briefing,  # TODO write a template to build a more robust instruction
            choices=list(self.toolboxes),
            k=k,
            max_validations=max_validations,
            system_message=system_message,
            **kwargs,
        )

    async def choose_tools(
        self,
        task: Task,
        toolbox: ToolBox,
        system_message: str = "",
        k: NonNegativeInt = 0,
        max_validations: PositiveInt = 2,
        **kwargs: Unpack[LLMKwargs],
    ) -> List[Tool]:
        """Asynchronously executes a multi-choice decision-making process to choose tools.

        Args:
            task (Task): The task for which to choose tools.
            toolbox (ToolBox): The toolbox from which to choose tools.
            system_message (str): Custom system-level prompt, defaults to an empty string.
            k (NonNegativeInt): The number of tools to select, 0 means infinite. Defaults to 0.
            max_validations (PositiveInt): Maximum number of validation failures, default is 2.
            **kwargs (Unpack[LLMKwargs]): Additional keyword arguments for the LLM usage.

        Returns:
            List[Tool]: The selected tools.
        """
        if not toolbox.tools:
            logger.warning(f"No tools available in toolbox {toolbox.name}.")
            return []
        return await self.achoose(
            instruction=task.briefing,  # TODO write a template to build a more robust instruction
            choices=toolbox.tools,
            k=k,
            max_validations=max_validations,
            system_message=system_message,
            **kwargs,
        )

    async def gather_tools_fine_grind(
        self,
        task: Task,
        box_choose_kwargs: Optional[ChooseKwargs] = None,
        tool_choose_kwargs: Optional[ChooseKwargs] = None,
    ) -> List[Tool]:
        """Asynchronously gathers tools based on the provided task and toolbox and tool selection criteria.

        Args:
            task (Task): The task for which to gather tools.
            box_choose_kwargs (Optional[ChooseKwargs]): Keyword arguments for choosing toolboxes.
            tool_choose_kwargs (Optional[ChooseKwargs]): Keyword arguments for choosing tools.

        Returns:
            List[Tool]: A list of tools gathered based on the provided task and toolbox and tool selection criteria.
        """
        box_choose_kwargs = box_choose_kwargs or {}
        tool_choose_kwargs = tool_choose_kwargs or {}

        # Choose the toolboxes
        chosen_toolboxes = await self.choose_toolboxes(task, **box_choose_kwargs)
        # Choose the tools
        chosen_tools = []
        for toolbox in chosen_toolboxes:
            chosen_tools.extend(await self.choose_tools(task, toolbox, **tool_choose_kwargs))
        return chosen_tools

    async def gather_tools(self, task: Task, **kwargs: Unpack[ChooseKwargs]) -> List[Tool]:
        """Asynchronously gathers tools based on the provided task.

        Args:
            task (Task): The task for which to gather tools.
            **kwargs (Unpack[ChooseKwargs]): Keyword arguments for choosing tools.

        Returns:
            List[Tool]: A list of tools gathered based on the provided task.
        """
        return await self.gather_tools_fine_grind(task, kwargs, kwargs)

    def supply_tools_from[S: "ToolBoxUsage"](self, others: Union[S, Iterable[S]]) -> Self:
        """Supplies tools from other ToolUsage instances to this instance.

        Args:
            others (ToolBoxUsage | Iterable[ToolBoxUsage]): A single ToolUsage instance or an iterable of ToolUsage instances
                from which to take tools.

        Returns:
            Self: The current ToolUsage instance with updated tools.
        """
        if isinstance(others, ToolBoxUsage):
            others = [others]
        for other in others:
            self.toolboxes.update(other.toolboxes)
        return self

    def provide_tools_to[S: "ToolBoxUsage"](self, others: Union[S, Iterable[S]]) -> Self:
        """Provides tools from this instance to other ToolUsage instances.

        Args:
            others (ToolBoxUsage | Iterable[ToolBoxUsage]): A single ToolUsage instance or an iterable of ToolUsage instances
                to which to provide tools.

        Returns:
            Self: The current ToolUsage instance.
        """
        if isinstance(others, ToolBoxUsage):
            others = [others]
        for other in others:
            other.toolboxes.update(self.toolboxes)
        return self



from pathlib import Path
from typing import Any, Dict, List, Optional

class TemplateManager:
    """TemplateManager class for managing handlebars templates."""
    def __init__(
        self, template_dirs: List[Path], suffix: Optional[str] = None, active_loading: Optional[bool] = None
    ) -> None:
        """Initialize the template manager.

        Args:
            template_dirs (List[Path]): A list of paths to directories containing templates.
            suffix (str, optional): The suffix of template files. None means 'hbs' suffix.
            active_loading (bool, optional): Whether to enable active loading of templates.
        """

    @property
    def template_count(self) -> int:
        """Get the number of templates discovered."""

    def get_template_source(self, name: str) -> Optional[str]:
        """Get the source path of a template by name.

        Args:
            name (str): The name of the template to retrieve.

        Returns:
            Optional[str]: The source path of the template.
        """

    def discover_templates(self) -> None:
        """Discover templates in the specified directories."""

    def render_template(self, name: str, data: Dict[str, Any]) -> str:
        """Render a template with the given name and data.

        Args:
            name (str): The name of the template to render.
            data (Dict[str, Any]): The data to pass to the template.

        Returns:
            str: The rendered template.
        """

def blake3_hash(content: bytes) -> str:
    """Calculate the BLAKE3 hash of the given data.

    Args:
        content (bytes): The data to hash.

    Returns:
        str: The BLAKE3 hash of the data.
    """
